# Copyright 2024 Matheus Vilano
# SPDX-License-Identifier: Apache-2.0

from waapi import WaapiClient as _WaapiClient

from pywwise.structs import WwiseObjectWatch
from pywwise.waapi.ak.wwise.console import Console as _Console
from pywwise.waapi.ak.wwise.core import Core as _Core
from pywwise.waapi.ak.wwise.debug import Debug as _Debug
from pywwise.waapi.ak.wwise.ui import UI as _UI
from pywwise.waapi.ak.wwise.waapi import Waapi as _Waapi


class Wwise:
    """ak.wwise"""
    
    def __init__(self, client: _WaapiClient, is_debug_build: bool = False, is_console_instance: bool = False,
                 watch_list: tuple[WwiseObjectWatch, ...] = ()):
        """
        Constructor.
        :param client: The WAAPI client to use.
        :param is_debug_build: Should be set to true if the instance of Wwise is a debug build and debug-only
                               functions/topics are required.
        :param is_console_instance: Should be set to true if the instance of Wwise is running in a console window.
        :param watch_list: A tuple of `WwiseObjectWatch` instances. This will be used to set up the
                           `ak.wwise.core.object.property_changed` event.
        """
        self._client = client
        self.console = _Console(client, is_console_instance)
        self.core = _Core(client, watch_list)
        self.debug = _Debug(client, is_debug_build)
        self.ui = _UI(client)
        self.waapi = _Waapi(client)



# Copyright 2025 Matheus Vilano
# SPDX-License-Identifier: Apache-2.0

from pywwise.objects.sources.impacter import Impacter
from pywwise.objects.sources.motion_source import MotionSource
from pywwise.objects.sources.soundseed_air_wind import SoundSeedAirWind
from pywwise.objects.sources.soundseed_air_woosh import SoundSeedAirWoosh
from pywwise.objects.sources.soundseed_grain import SoundSeedGrain
from pywwise.objects.sources.wwise_silence import WwiseSilence
from pywwise.objects.sources.wwise_synth_one import WwiseSynthOne
from pywwise.objects.sources.wwise_tone_generator import WwiseToneGenerator



# Copyright 2025 Matheus Vilano
# SPDX-License-Identifier: Apache-2.0

from pywwise.descriptors import WwiseProperty
from pywwise.enums import EToneGenChannelMask, EToneGenDurationMode, EToneGenSweepFrequencyType, EToneGenWaveform
from pywwise.objects.abc import WwiseObject


class WwiseToneGenerator(WwiseObject):
    """
    https://www.audiokinetic.com/en/library/edge/?source=SDK&id=wwiseobject_source_wwise_tone_generator.html \n
    A class serving as an interface for getting/setting properties on Wwise objects. This type specifically targets
    the class represented by `EObjectType.SOURCE_PLUGIN`.
    """
    attack_time = WwiseProperty[float]("AttackTime", float)
    channel_mask = WwiseProperty[EToneGenChannelMask]("ChannelMask", EToneGenChannelMask)
    decay_time = WwiseProperty[float]("DecayTime", float)
    duration_mode = WwiseProperty[EToneGenDurationMode]("DurMode", EToneGenDurationMode)
    release_time = WwiseProperty[float]("ReleaseTime", float)
    start_frequency = WwiseProperty[float]("StartFreq", float)
    start_frequency_random_max = WwiseProperty[float]("StartFreqRandMax", float)
    start_frequency_random_min = WwiseProperty[float]("StartFreqRandMin", float)
    stop_frequency = WwiseProperty[float]("StopFreq", float)
    stop_frequency_random_max = WwiseProperty[float]("StopFreqRandMax", float)
    stop_frequency_random_min = WwiseProperty[float]("StopFreqRandMin", float)
    sustain_level = WwiseProperty[float]("SustainLevel", float)
    sustian_time = WwiseProperty[float]("SustainTime", float)
    sweep_frequency = WwiseProperty[float]("SweepFreq", float)
    sweep_frequency_type = WwiseProperty[EToneGenSweepFrequencyType]("SweepFreqType", EToneGenSweepFrequencyType)
    wave_gain = WwiseProperty[float]("WaveGain", float)
    wave_type = WwiseProperty[EToneGenWaveform]("WaveType", EToneGenWaveform)



# Copyright 2025 Matheus Vilano
# SPDX-License-Identifier: Apache-2.0

from pywwise.descriptors import WwiseProperty
from pywwise.enums import EMatrixReverbDelayCount, EMatrixReverbDelayLengthsMode
from pywwise.objects.abc import WwiseObject


class WwiseMatrixReverb(WwiseObject):
    """
    https://www.audiokinetic.com/en/library/edge/?source=SDK&id=wwiseobject_effect_wwise_matrix_reverb.html \n
    A class serving as an interface for getting/setting properties on Wwise objects. This type specifically targets
    the class represented by `EObjectType.WWISE_MATRIX_REVERB`.
    """
    delay_lengths_mode = WwiseProperty[EMatrixReverbDelayLengthsMode]("DelayLengthsMode", EMatrixReverbDelayLengthsMode)
    delay_time_1 = WwiseProperty[float]("DelayTime1", float)
    delay_time_2 = WwiseProperty[float]("DelayTime2", float)
    delay_time_3 = WwiseProperty[float]("DelayTime3", float)
    delay_time_4 = WwiseProperty[float]("DelayTime4", float)
    delay_time_5 = WwiseProperty[float]("DelayTime5", float)
    delay_time_6 = WwiseProperty[float]("DelayTime6", float)
    delay_time_7 = WwiseProperty[float]("DelayTime7", float)
    delay_time_8 = WwiseProperty[float]("DelayTime8", float)
    delay_time_9 = WwiseProperty[float]("DelayTime9", float)
    delay_time_10 = WwiseProperty[float]("DelayTime10", float)
    delay_time_11 = WwiseProperty[float]("DelayTime11", float)
    delay_time_12 = WwiseProperty[float]("DelayTime12", float)
    delay_time_13 = WwiseProperty[float]("DelayTime13", float)
    delay_time_14 = WwiseProperty[float]("DelayTime14", float)
    delay_time_15 = WwiseProperty[float]("DelayTime15", float)
    delay_time_16 = WwiseProperty[float]("DelayTime16", float)
    dry_level = WwiseProperty[float]("DryLevel", float)
    hf_ratio = WwiseProperty[float]("HFRatio", float)
    number_of_delays = WwiseProperty[EMatrixReverbDelayCount]("NumberOfDelays", EMatrixReverbDelayCount)
    pre_delay = WwiseProperty[float]("PreDelay", float)
    process_lfe = WwiseProperty[bool]("ProcessLFE", bool)
    reverb_time = WwiseProperty[float]("ReverbTime", float)
    wet_level = WwiseProperty[float]("WetLevel", float)



# Copyright 2024 Matheus Vilano
# SPDX-License-Identifier: Apache-2.0

from waapi import WaapiClient as _WaapiClient

from pywwise.aliases import ListOrTuple, SystemPath
from pywwise.enums import EReturnOptions, ESourceControlSearchFilter, ESourceFileReturnOptions
from pywwise.primitives import OriginalsPath
from pywwise.structs import LogItem, SourceControlStatus, SourceFileInfo, WwiseObjectInfo


class SourceControl:
    """ak.wwise.core.sourceControl"""
    
    def __init__(self, client: _WaapiClient):
        """
        Constructor.
        :param client: The WAAPI client to use.
        """
        self._client = client
    
    def add(self, files: ListOrTuple[SystemPath]) -> tuple[LogItem, ...]:
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_wwise_core_sourcecontrol_add.html \n
        Add files to source control. Equivalent to Mark for Add for Perforce.
        :param files: Array of files to add. File paths must be absolute.
        :return: A tuple containing log items generating during the request.
        """
        files = list(dict.fromkeys(files))  # convert to list of unique items
        result = self._client.call("ak.wwise.core.sourceControl.add", {"files": files})
        return tuple([LogItem.from_dict(result["log"])] if result is not None else [])
    
    def check_out(self, files: ListOrTuple[SystemPath]) -> tuple[LogItem, ...]:
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_wwise_core_sourcecontrol_checkout.html \n
        Check out files from source control. Equivalent to Check Out for Perforce.
        :param files: Array of files to check out. File paths must be absolute.
        :return: A tuple containing log items generating during the request.
        """
        files = list(dict.fromkeys(files))  # convert to list of unique items
        result = self._client.call("ak.wwise.core.sourceControl.checkOut", {"files": files})
        return tuple([LogItem.from_dict(result["log"])] if result is not None else [])
    
    def commit(self, files: ListOrTuple[SystemPath], message: str) -> tuple[LogItem, ...]:
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_wwise_core_sourcecontrol_commit.html \n
        Commit files to source control. Equivalent to Submit Changes for Perforce.
        :param files: Array of files to commit. File paths must be absolute.
        :param message: Description message for the commit.
        :return: A tuple containing log items generating during the request.
        """
        files = list(dict.fromkeys(files))  # convert to list of unique items
        result = self._client.call("ak.wwise.core.sourceControl.commit", {"files": files, "message": message})
        return tuple([LogItem.from_dict(result["log"])] if result is not None else [])
    
    def delete(self, files: ListOrTuple[SystemPath]) -> tuple[LogItem, ...]:
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_wwise_core_sourcecontrol_delete.html \n
        Delete files from source control. Equivalent to Mark for Delete for Perforce.
        :param files: Array of files to delete. File paths must be absolute.
        :return: A tuple containing log items generating during the request.
        """
        files = list(dict.fromkeys(files))  # convert to list of unique items
        result = self._client.call("ak.wwise.core.sourceControl.delete", {"files": files})
        return tuple([LogItem.from_dict(result["log"])] if result is not None else [])
    
    def get_source_files(self, folder: str = None, recursive: bool = True,
                         search_filter: ESourceControlSearchFilter = ESourceControlSearchFilter.ALL) -> tuple[
        SourceFileInfo, ...]:
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_wwise_core_sourcecontrol_getsourcefiles.html \n
        Retrieve all original files.
        :param folder: Base folder for search relative to Originals folder. Default to the Originals folder.
        :param recursive: Search in all subfolders of the base folder.
        :param search_filter: Filter the files are in the search result.
        :return: A tuple containing SourceFileInfo instances, which represent information on each source file found.
                 Empty is no files were found.
        """
        args = {"recursive": recursive, "filter": search_filter}
        if folder is not None:
            args["folder"] = folder
        
        options = {"return": [ESourceFileReturnOptions.FILE, ESourceFileReturnOptions.USAGE,
                              ESourceFileReturnOptions.IS_MISSING],
                   "objectReturn": EReturnOptions.get_defaults()}
        
        results = self._client.call("ak.wwise.core.sourceControl.getSourceFiles", args, options=options)
        if results is None:
            return ()
        
        returns = list[SourceFileInfo]()
        for result in results.get("return", ()):
            file = OriginalsPath(result["file"])
            usage = tuple([WwiseObjectInfo.from_dict(obj) for obj in result.get("usage", ())])
            is_missing = result["isMissing"]
            returns.append(SourceFileInfo(file, usage, is_missing))
        return tuple(returns)
    
    def get_status(self, files: ListOrTuple[tuple[SystemPath, SystemPath]]) -> tuple[
        tuple[LogItem, ...], tuple[SourceControlStatus, ...]]:
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_wwise_core_sourcecontrol_getstatus.html \n
        Get the source control status of the specified files.
        :param files: Array of files for which to retrieve source control status. File paths must be absolute.
        :return: Two tuples: the first one containing log items generating during the request; the second one
                 containing the source control status for each file passed to `files`.
        """
        files = list(dict.fromkeys(files))  # convert to list of unique items
        result = self._client.call("ak.wwise.core.sourceControl.delete", {"files": files})
        if result is None:
            return (), ()
        logs = tuple([LogItem.from_dict(result["log"])])
        statuses = tuple([SourceControlStatus(result.get("result", dict()).get("status", ""),
                                              result.get("result", dict()).get("owner", ""))])
        return logs, statuses
    
    def move(self, files: ListOrTuple[tuple[SystemPath, SystemPath]]) -> tuple[LogItem, ...]:
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_wwise_core_sourcecontrol_move.html \n
        Move or rename files in source control. Always pass the same number of elements in files and
        newFiles. Equivalent to Move for Perforce.
        :param files: Array of files to move. File paths must be absolute.
        :return: A tuple containing log items generating during the request.
        """
        original_paths, new_paths = zip(*files)
        args = {"files": original_paths, "newFiles": new_paths}
        result = self._client.call("ak.wwise.core.sourceControl.move", args)
        return tuple([LogItem.from_dict(result["log"])] if result is not None else [])
    
    def revert(self, files: ListOrTuple[SystemPath]) -> tuple[LogItem, ...]:
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_wwise_core_sourcecontrol_revert.html \n
        Revert changes to files in source control.
        :param files: Array of files to revert. File paths must be absolute.
        :return: A tuple containing log items generating during the request.
        """
        files = list(dict.fromkeys(files))  # convert to list of unique items
        result = self._client.call("ak.wwise.core.sourceControl.revert", {"files": files})
        return tuple([LogItem.from_dict(result["log"])] if result is not None else [])
    
    def set_provider(self, provider: str, server: str = "localhost", port: str = "", username: str = "",
                     password: str = "", workspace: str = "", host: str = "") -> tuple[LogItem, ...]:
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_wwise_core_sourcecontrol_setprovider.html \n
        Change the source control provider and credentials. This is the same setting as the Source
        Control option in the Project Settings dialog in Wwise.
        :param provider: Source control provider. Allowed values "Perforce", "Subversion", any other source control
                         supported or an empty string. An empty value is the equivalent of choosing No Control Source
                         in the UI.
        :param server: Server name or address. Default value is localhost.
        :param port: Server port. This value is required for Perforce.
        :param username: Username. This value is required for Perforce.
        :param password: User password or ticket. This value is required for Perforce.
        :param workspace: Workspace name. This value is required for Perforce.
        :param host: Host name. This value is required for Perforce.
        :return: A tuple containing log items generating during the request.
        """
        args = {"provider": provider, "server": server, "port": port, "username": username, "password": password,
                "workspace": workspace, "host": host}
        result = self._client.call("ak.wwise.core.sourceControl.setProvider", args)
        return tuple([LogItem.from_dict(result["log"])] if result is not None else [])



# Copyright 2025 Matheus Vilano
# SPDX-License-Identifier: Apache-2.0

from pywwise.descriptors import WwiseProperty
from pywwise.enums import EModPhaseMode, EModWaveform
from pywwise.objects.abc import WwiseObject


class WwiseTremolo(WwiseObject):
    """
     \n
    A class serving as an interface for getting/setting properties on Wwise objects. This type specifically targets
    the class represented by `EObjectType.WWISE_TREMOLO`.
    """
    mod_depth = WwiseProperty[float]("ModDepth", float)
    mod_frequency = WwiseProperty[float]("ModFrequency", float)
    mod_pwm = WwiseProperty[float]("ModPWM", float)
    mod_phase_mode = WwiseProperty[EModPhaseMode]("ModPhaseMode", EModPhaseMode)
    mod_phase_offset = WwiseProperty[float]("ModPhaseOffset", float)
    mod_phase_spread = WwiseProperty[float]("ModPhaseSpread", float)
    mod_smoothing = WwiseProperty[float]("ModSmoothing", float)
    mod_waveform = WwiseProperty[EModWaveform]("ModWaveform", EModWaveform)
    output_gain = WwiseProperty[float]("OutputGain", float)
    process_centrering = WwiseProperty[bool]("ProcessCenter", bool)
    process_lfe = WwiseProperty[bool]("ProcessLFE", bool)



# Copyright 2024 Matheus Vilano
# SPDX-License-Identifier: Apache-2.0

from base64 import b64decode as _b64decode
from datetime import datetime as _datetime
from os import makedirs as _makedirs

from simplevent import RefEvent as _RefEvent
from waapi import WaapiClient as _WaapiClient

from pywwise.aliases import ListOrTuple, SystemPath
from pywwise.decorators import callback
from pywwise.enums import EObjectType, EReturnOptions
from pywwise.primitives import GUID, Name, ProjectPath
from pywwise.structs import Rect, WwiseObjectInfo
from pywwise.waapi.ak.wwise.ui.commands import Commands as _Commands
from pywwise.waapi.ak.wwise.ui.project import Project as _Project


class UI:
    """ak.wwise.ui"""
    
    def __init__(self, client: _WaapiClient):
        """
        Constructor.
        :param client: The WAAPI client to use.
        """
        self._client = client
        
        self.commands = _Commands(client)
        self.project = _Project(client)
        
        self.selection_changed = _RefEvent(tuple[WwiseObjectInfo, ...])
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_wwise_ui_selectionchanged.html
        \nSent when the selection changes in the project.
        \n**Event Data**:
        \n- A tuple of WwiseObjectInfo instances (each containing a GUID, a name, a type, and a path).
        """
        
        selection_changed_options = {"return": [EReturnOptions.GUID, EReturnOptions.NAME,
                                                EReturnOptions.TYPE, EReturnOptions.PATH]}
        self._selection_changed = self._client.subscribe("ak.wwise.ui.selectionChanged", self._on_selection_changed,
                                                         selection_changed_options)
    
    @callback
    def _on_selection_changed(self, event: _RefEvent, **kwargs):
        """
        Callback function for the `selectionChanged` event.
        :param event: The event to broadcast.
        :param kwargs: The event data.
        """
        objects = list[WwiseObjectInfo]()
        for obj in kwargs["objects"]:
            objects.append(WwiseObjectInfo.from_dict(obj))
        event(tuple(objects))
    
    def bring_to_foreground(self) -> bool:
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_wwise_ui_bringtoforeground.html \n
        Bring Wwise main window to foreground. Refer to `SetForegroundWindow` and `AllowSetForegroundWindow`
        on MSDN for more information on the restrictions. Refer to `ak.wwise.core.get_info` to obtain the
        Wwise process ID for `AllowSetForegroundWindow`.
        :return: Whether the call succeeded.
        """
        return self._client.call("ak.wwise.ui.bringToForeground") is not None
    
    def capture_screen(self, view_name: str = None, view_selection_channel: int = None, capture_rect: Rect = None,
                       output_path: SystemPath = None) -> tuple[str, str]:
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_wwise_ui_capturescreen.html \n
        Captures a part of the Wwise UI relative to a view.
        :param view_name: The name of the view as displayed in Wwise UI. By default, the whole UI is captured.
        :param view_selection_channel: The selection channel of the view. Can be a value of 1, 2, 3 or 4. By default,
                                       the current selection channel of the view is detected automatically. If the
                                       specified value is out of bounds, the value will be clamped.
        :param capture_rect: The capture region. By default, the whole view is captured.
        :param output_path: If specified, a PNG image will be created at the specified location. If only directory
                            names are specified (e.g. "C:/Users/PyWwise/Pictures"), the file name will be the current
                            system date and time.
        :return: The underlying image data format and the encoded image data (Base64).
        """
        args = {}
        
        if view_name is not None:
            args["viewName"] = view_name
        if view_selection_channel is not None:
            args["viewSelectionChannel"] = max(min(view_selection_channel, 4), 1)
        if capture_rect is not None:
            args["rect"] = {"x": capture_rect.x, "y": capture_rect.y, "width": capture_rect.width,
                            "height": capture_rect.height}
        
        results = self._client.call("ak.wwise.ui.captureScreen", args)
        content_type, content_base = results.get("contentType"), results.get("contentBase64")
        
        if output_path is not None:
            if output_path.suffix == "":
                date_time = _datetime.now()
                date = f"{date_time.year}-{date_time.month:02}-{date_time.day:02}"
                time = f"{date_time.hour:02}{date_time.minute:02}{date_time.second:02}"
                output_path /= f"Screenshot {date} {time}.png"
            if not output_path.exists():
                _makedirs(output_path.parent, exist_ok=True)
            with open(output_path, "wb") as image:
                image.write(_b64decode(content_base))
        
        return content_type, content_base
    
    def get_selected_objects(self, returns: ListOrTuple[EReturnOptions] = None, platform: str = None,
                             language: str = None) -> tuple[WwiseObjectInfo, ...]:
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_wwise_ui_getselectedobjects.html \n
        Retrieves the list of objects currently selected by the user in the active view.
        :param returns: The additional return options. By default, this function returns only the GUID and Name
                        of the selected objects. Duplicates are ignored.
        :param platform: If specified, this function will get information from the specified platform instead of the
                         current platform.
        :param language: If specified, this function will get information from the specified language instead of the
                         current language.
        :returns: For each selected object, a WwiseObjectInfo containing an object's GUID, name, path, type, and a
                  dictionary containing additional information (requested via `return_options`). When accessing the
                  values in the dictionary, use the EReturnOptions enum as the keys. If this function call fails, an
                  empty tuple is returned.
        """
        returns = list(dict.fromkeys(returns)) if returns is not None else list[EReturnOptions]()
        returns.extend(EReturnOptions.get_defaults())
        options = {"return": returns}  # to ensure only unique values
        
        if platform is not None:
            options["platform"] = platform
        if language is not None:
            options["language"] = language
        
        results = self._client.call("ak.wwise.ui.getSelectedObjects", dict(), options=options)
        results = results.get("objects") if results is not None else None
        if results is None:
            return ()
        
        objects = list[WwiseObjectInfo]()
        
        for result in results:
            guid = GUID(result[EReturnOptions.GUID])
            name = Name(result[EReturnOptions.NAME])
            typename = EObjectType.from_type_name(result[EReturnOptions.TYPE])
            path = ProjectPath(result[EReturnOptions.PATH])
            other = {key: value for key, value in result.items() if key not in EReturnOptions.get_defaults()}
            objects.append(WwiseObjectInfo(guid, name, typename, path, other))
        
        return tuple(objects)



# Copyright 2025 Matheus Vilano
# SPDX-License-Identifier: Apache-2.0

from pywwise.descriptors import WwiseProperty
from pywwise.enums import EActionNamePosition, ECaseStyleSimple, ESoundBankDefinitionFormat
from pywwise.objects.abc import WwiseObject
from pywwise.objects.types.conversion import Conversion
from pywwise.primitives import ProjectPath


class UserProjectSettings(WwiseObject):
    """
    https://www.audiokinetic.com/en/library/edge/?source=SDK&id=wwiseobject_userprojectsettings.html \n
    A class serving as an interface for getting/setting properties on Wwise objects. This type specifically targets
    the class represented by `EObjectType.USER_PROJECT_SETTINGS`.
    """
    auto_sound_bank_all_events = WwiseProperty[bool]("AutoSoundBankAllEvents", bool)
    auto_sound_bank_enabled = WwiseProperty[bool]("AutoSoundBankEnabled", bool)
    conversion = WwiseProperty[Conversion]("Conversion", Conversion)
    convert_external_sources = WwiseProperty[bool]("ConvertExternalSources", bool)
    copy_loose_streamed_media = WwiseProperty[bool]("CopyLooseStreamedMedia", bool)
    default_sound_volume = WwiseProperty[float]("DefaultSoundVolume", float)
    event_action_name_position = WwiseProperty[EActionNamePosition]("EventActionNamePosition", EActionNamePosition)
    event_creation_settings_override = WwiseProperty[bool]("EventCreationSettingsOverride", bool)
    event_name_case_type = WwiseProperty[ECaseStyleSimple]("EventNameCaseType", ECaseStyleSimple)
    event_name_modify_case = WwiseProperty[bool]("EventNameModifyCase", bool)
    external_sources_input_path = WwiseProperty[ProjectPath]("ExternalSourcesInputPath", ProjectPath)
    external_sources_output_path = WwiseProperty[ProjectPath]("ExternalSourcesOutputPath", ProjectPath)
    generate_main_sound_bank = WwiseProperty[bool]("GenerateMainSoundBank", bool)
    generate_multiple_banks = WwiseProperty[bool]("GenerateMultipleBanks", bool)
    generate_sound_bank_json = WwiseProperty[bool]("GenerateSoundBankJSON", bool)
    generate_sound_bank_xml = WwiseProperty[bool]("GenerateSoundBankXML", bool)
    media_auto_bank_sub_folders = WwiseProperty[bool]("MediaAutoBankSubFolders", bool)
    override_conversion = WwiseProperty[bool]("OverrideConversion", bool)
    post_generate_step_user_override = WwiseProperty[bool]("PostGenerateStepUserOverride", bool)
    pre_generate_step_user_override = WwiseProperty[bool]("PreGenerateStepUserOverride", bool)
    remove_unused_generated_files = WwiseProperty[bool]("RemoveUnusedGeneratedFiles", bool)
    settings_user_override = WwiseProperty[bool]("SettingsUserOverride", bool)
    sound_bank_allow_exceeding_sb = WwiseProperty[bool]("SoundBankAllowExceedingSB", bool)
    sound_bank_definition_file_format = WwiseProperty[ESoundBankDefinitionFormat]("SoundBankDefinitionFileFormat",
                                                                                  ESoundBankDefinitionFormat)
    sound_bank_generate_definition_file = WwiseProperty[bool]("SoundBankGenerateDefinitionFile", bool)
    sound_bank_generate_estimated_duration = WwiseProperty[bool]("SoundBankGenerateEstimatedDuration", bool)
    sound_bank_generate_header_file = WwiseProperty[bool]("SoundBankGenerateHeaderFile", bool)
    sound_bank_generate_max_attenuation_info = WwiseProperty[bool]("SoundBankGenerateMaxAttenuationInfo", bool)
    sound_bank_generate_print_colour = WwiseProperty[bool]("SoundBankGeneratePrintColor", bool)
    sound_bank_generate_print_guid = WwiseProperty[bool]("SoundBankGeneratePrintGUID", bool)
    sound_bank_generate_print_path = WwiseProperty[bool]("SoundBankGeneratePrintPath", bool)
    sound_bank_generate_readable_file = WwiseProperty[bool]("SoundBankGenerateReadableFile", bool)
    sound_bank_header_file_path = WwiseProperty[ProjectPath]("SoundBankHeaderFilePath", ProjectPath)
    sound_bank_include_soundbank_names_strings = WwiseProperty[bool]("SoundBankIncludeSoundbankNamesStrings", bool)
    sound_bank_paths = WwiseProperty[str]("SoundBankPaths", str)
    sound_bank_post_generate_custom_cmd_description = WwiseProperty[str]("SoundBankPostGenerateCustomCmdDescription",
                                                                         str)
    sound_bank_post_generate_custom_cmd_lines = WwiseProperty[str]("SoundBankPostGenerateCustomCmdLines", str)
    sound_bank_pre_generate_custom_cmd_description = WwiseProperty[str]("SoundBankPreGenerateCustomCmdDescription", str)
    sound_bank_pre_generate_custom_cmd_lines = WwiseProperty[str]("SoundBankPreGenerateCustomCmdLines", str)
    sound_bank_update_audio_files = WwiseProperty[bool]("SoundBankUpdateAudioFiles", bool)
    source_control_generated_files = WwiseProperty[bool]("SourceControlGeneratedFiles", bool)
    use_action_name_for_event = WwiseProperty[bool]("UseActionNameForEvent", bool)



# Copyright 2025 Matheus Vilano
# SPDX-License-Identifier: Apache-2.0

from pywwise.descriptors import WwiseProperty
from pywwise.enums import EHarmonizerFilterType, EHarmonizerInput
from pywwise.objects.abc import WwiseObject


class WwiseHarmonizer(WwiseObject):
    """
    https://www.audiokinetic.com/en/library/edge/?source=SDK&id=wwiseobject_effect_wwise_harmonizer.html \n
    A class serving as an interface for getting/setting properties on Wwise objects. This type specifically targets
    the class represented by `EObjectType.WWISE_HARMONIZER`.
    """
    dry_level = WwiseProperty[float]("DryLevel", float)
    input_format = WwiseProperty[EHarmonizerInput]("Input", EHarmonizerInput)
    process_lfe = WwiseProperty[bool]("ProcessLFE", bool)
    sync_dry = WwiseProperty[bool]("SyncDry", bool)
    voice1_enable = WwiseProperty[bool]("Voice1Enable", bool)
    voice1_filter_frequency = WwiseProperty[float]("Voice1FilterFrequency", float)
    voice1_filter_gain = WwiseProperty[float]("Voice1FilterGain", float)
    voice1_filter_q = WwiseProperty[float]("Voice1FilterQFactor", float)
    voice1_filter_type = WwiseProperty[EHarmonizerFilterType]("Voice1FilterType", EHarmonizerFilterType)
    voice1_gain = WwiseProperty[float]("Voice1Gain", float)
    voice1_pitch = WwiseProperty[float]("Voice1Pitch", float)
    voice2_enable = WwiseProperty[bool]("Voice2Enable", bool)
    voice2_filter_frequency = WwiseProperty[float]("Voice2FilterFrequency", float)
    voice2_filter_gain = WwiseProperty[float]("Voice2FilterGain", float)
    voice2_filter_q = WwiseProperty[float]("Voice2FilterQFactor", float)
    voice2_filter_type = WwiseProperty[EHarmonizerFilterType]("Voice2FilterType", EHarmonizerFilterType)
    voice2_gain = WwiseProperty[float]("Voice2Gain", float)
    voice2_pitch = WwiseProperty[float]("Voice2Pitch", float)
    wet_level = WwiseProperty[float]("WetLevel", float)
    window_size = WwiseProperty[int]("WindowSize", int)



# Copyright 2025 Matheus Vilano
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

from typing import TYPE_CHECKING, Union as _Union

if TYPE_CHECKING:
    from pywwise.objects.types.music_stinger import MusicStinger

from pywwise.descriptors import WwiseProperty
from pywwise.enums import (E3DPosition, E3DSpatialization, EColour, EDiscardBehaviour, ELoudnessNormalizationType,
                           EMidiTempoSource, EMusicalDuration, EMusicalGridFrequency, EOverLimitBehaviour, EScope,
                           ESpeakerPanning, ETimeSignature, EVirtualVoiceBehaviour, EVirtualVoiceQueueBehaviour)
from pywwise.objects.abc import WwiseObject
from pywwise.objects.types.attenuation import Attenuation
from pywwise.objects.types.aux_bus import AuxBus
from pywwise.objects.types.blend_container import BlendContainer
from pywwise.objects.types.bus import Bus
from pywwise.objects.types.conversion import Conversion
from pywwise.objects.types.effect_slot import EffectSlot
from pywwise.objects.types.music_cue import MusicCue
from pywwise.objects.types.random_sequence_container import RandomSequenceContainer
from pywwise.objects.types.rtpc import Rtpc
from pywwise.objects.types.sound import Sound
from pywwise.objects.types.switch_container import SwitchContainer
from pywwise.primitives import GUID

_MidiTargetType = _Union[RandomSequenceContainer, SwitchContainer, BlendContainer, Sound]

_MidiTargetTypeTuple = (RandomSequenceContainer, SwitchContainer, BlendContainer, Sound)


class MusicSegment(WwiseObject):
    """
    https://www.audiokinetic.com/en/library/edge/?source=SDK&id=wwiseobject_musicsegment.html \n
    A class serving as an interface for getting/setting properties on Wwise objects. This type specifically targets
    the class represented by `EObjectType.MUSIC_SEGMENT`.
    """
    position_3d = WwiseProperty[E3DPosition]("3DPosition", E3DPosition)
    spatialization_3d = WwiseProperty[E3DSpatialization]("3DSpatialization", E3DSpatialization)
    attenuation = WwiseProperty[Attenuation]("Attenuation", Attenuation)
    attenuation_distance_scaling = WwiseProperty[float]("AttenuationDistanceScaling", float)
    below_threshold_behaviour = WwiseProperty[EVirtualVoiceBehaviour]("BelowThresholdBehavior", EVirtualVoiceBehaviour)
    bypass_effect = WwiseProperty[bool]("BypassEffect", bool)
    centre_percentage = WwiseProperty[float]("CenterPercentage", float)
    colour = WwiseProperty[EColour]("Color", EColour)
    conversion = WwiseProperty[Conversion]("Conversion", Conversion)
    cues = WwiseProperty[tuple[MusicCue, ...]]("Cues", tuple)
    effects = WwiseProperty[tuple[EffectSlot, ...]]("Effects", tuple)
    enable_attenuation = WwiseProperty[bool]("EnableAttenuation", bool)
    enable_diffraction = WwiseProperty[bool]("EnableDiffraction", bool)
    enable_loudness_normalization = WwiseProperty[bool]("EnableLoudnessNormalization", bool)
    end_position = WwiseProperty[float]("EndPosition", float)
    game_aux_send_hpf = WwiseProperty[int]("GameAuxSendHPF", int)
    game_aux_send_lpf = WwiseProperty[int]("GameAuxSendLPF", int)
    game_aux_send_volume = WwiseProperty[float]("GameAuxSendVolume", float)
    grid_frequency_preset = WwiseProperty[EMusicalGridFrequency]("GridFrequencyPreset", EMusicalGridFrequency)
    grid_offset_custom = WwiseProperty[float]("GridOffsetCustom", float)
    grid_offset_preset = WwiseProperty[EMusicalDuration]("GridOffsetPreset", EMusicalDuration)
    hdr_active_range = WwiseProperty[float]("HdrActiveRange", float)
    hdr_enable_envelope = WwiseProperty[bool]("HdrEnableEnvelope", bool)
    hdr_envelope_sensitivity = WwiseProperty[float]("HdrEnvelopeSensitivity", float)
    highpass = WwiseProperty[int]("Highpass", int)
    hold_emitter_position_orientation = WwiseProperty[bool]("HoldEmitterPositionOrientation", bool)
    hold_listener_orientation = WwiseProperty[bool]("HoldListenerOrientation", bool)
    ignore_parent_max_sound_instance = WwiseProperty[bool]("IgnoreParentMaxSoundInstance", bool)
    inclusion = WwiseProperty[bool]("Inclusion", bool)
    is_global_limit = WwiseProperty[EScope]("IsGlobalLimit", EScope)
    listener_relative_routing = WwiseProperty[bool]("ListenerRelativeRouting", bool)
    loudness_normalization_target = WwiseProperty[float]("LoudnessNormalizationTarget", float)
    loudness_normalization_type = WwiseProperty[ELoudnessNormalizationType](
        "LoudnessNormalizationType", ELoudnessNormalizationType)
    lowpass = WwiseProperty[int]("Lowpass", int)
    make_up_gain = WwiseProperty[float]("MakeUpGain", float)
    max_reached_behaviour = WwiseProperty[EDiscardBehaviour]("MaxReachedBehavior", EDiscardBehaviour)
    max_sound_per_instance = WwiseProperty[int]("MaxSoundPerInstance", int)
    metadata = WwiseProperty[GUID]("Metadata", GUID)
    midi_target = WwiseProperty[_MidiTargetType]("MidiTarget", _MidiTargetTypeTuple)
    midi_tempo_source = WwiseProperty[EMidiTempoSource]("MidiTempoSource", EMidiTempoSource)
    output_bus = WwiseProperty[Bus]("OutputBus", Bus)
    output_bus_highpass = WwiseProperty[int]("OutputBusHighpass", int)
    output_bus_lowpass = WwiseProperty[int]("OutputBusLowpass", int)
    output_bus_volume = WwiseProperty[float]("OutputBusVolume", float)
    over_limit_behaviour = WwiseProperty[EOverLimitBehaviour]("OverLimitBehavior", EOverLimitBehaviour)
    override_analysis = WwiseProperty[bool]("OverrideAnalysis", bool)
    override_clock_settings = WwiseProperty[bool]("OverrideClockSettings", bool)
    override_colour = WwiseProperty[bool]("OverrideColor", bool)
    override_conversion = WwiseProperty[bool]("OverrideConversion", bool)
    override_early_reflections = WwiseProperty[bool]("OverrideEarlyReflections", bool)
    override_effect = WwiseProperty[bool]("OverrideEffect", bool)
    override_game_aux_sends = WwiseProperty[bool]("OverrideGameAuxSends", bool)
    override_hdr_envelope = WwiseProperty[bool]("OverrideHdrEnvelope", bool)
    override_metadata = WwiseProperty[bool]("OverrideMetadata", bool)
    override_midi_target = WwiseProperty[bool]("OverrideMidiTarget", bool)
    override_midi_tempo = WwiseProperty[bool]("OverrideMidiTempo", bool)
    override_output = WwiseProperty[bool]("OverrideOutput", bool)
    override_positioning = WwiseProperty[bool]("OverridePositioning", bool)
    override_priority = WwiseProperty[bool]("OverridePriority", bool)
    override_user_aux_sends = WwiseProperty[bool]("OverrideUserAuxSends", bool)
    override_virtual_voice = WwiseProperty[bool]("OverrideVirtualVoice", bool)
    playback_speed = WwiseProperty[float]("PlaybackSpeed", float)
    priority = WwiseProperty[int]("Priority", int)
    priority_distance_factor = WwiseProperty[bool]("PriorityDistanceFactor", bool)
    priority_distance_offset = WwiseProperty[int]("PriorityDistanceOffset", int)
    rtpc = WwiseProperty[tuple[Rtpc, ...]]("RTPC", tuple)
    reflections_aux_send = WwiseProperty[AuxBus]("ReflectionsAuxSend", AuxBus)
    reflections_volume = WwiseProperty[float]("ReflectionsVolume", float)
    speaker_panning = WwiseProperty[ESpeakerPanning]("SpeakerPanning", ESpeakerPanning)
    speaker_panning_3d_spatialization_mix = WwiseProperty[int]("SpeakerPanning3DSpatializationMix", int)
    stingers: WwiseProperty[tuple[MusicStinger, ...]] = WwiseProperty(
        "Stingers", tuple)  # Using `:` to avoid circular imports.
    tempo = WwiseProperty[float]("Tempo", float)
    time_signature_lower = WwiseProperty[ETimeSignature]("TimeSignatureLower", ETimeSignature)
    timer_signature_upper = WwiseProperty[int]("TimeSignatureUpper", int)
    use_game_aux_sends = WwiseProperty[bool]("UseGameAuxSends", bool)
    use_max_sound_per_instance = WwiseProperty[bool]("UseMaxSoundPerInstance", bool)
    user_aux_send_0 = WwiseProperty[AuxBus]("UserAuxSend0", AuxBus)
    user_aux_send_1 = WwiseProperty[AuxBus]("UserAuxSend1", AuxBus)
    user_aux_send_2 = WwiseProperty[AuxBus]("UserAuxSend2", AuxBus)
    user_aux_send_3 = WwiseProperty[AuxBus]("UserAuxSend3", AuxBus)
    user_aux_send_hpf_0 = WwiseProperty[int]("UserAuxSendHPF0", int)
    user_aux_send_hpf_1 = WwiseProperty[int]("UserAuxSendHPF1", int)
    user_aux_send_hpf_2 = WwiseProperty[int]("UserAuxSendHPF2", int)
    user_aux_send_hpf_3 = WwiseProperty[int]("UserAuxSendHPF3", int)
    user_aux_send_lpf_0 = WwiseProperty[int]("UserAuxSendLPF0", int)
    user_aux_send_lpf_1 = WwiseProperty[int]("UserAuxSendLPF1", int)
    user_aux_send_lpf_2 = WwiseProperty[int]("UserAuxSendLPF2", int)
    user_aux_send_lpf_3 = WwiseProperty[int]("UserAuxSendLPF3", int)
    user_aux_send_volume_0 = WwiseProperty[float]("UserAuxSendVolume0", float)
    user_aux_send_volume_1 = WwiseProperty[float]("UserAuxSendVolume1", float)
    user_aux_send_volume_2 = WwiseProperty[float]("UserAuxSendVolume2", float)
    user_aux_send_volume_3 = WwiseProperty[float]("UserAuxSendVolume3", float)
    virtual_voice_queue_behaviour = WwiseProperty[EVirtualVoiceQueueBehaviour](
        "VirtualVoiceQueueBehavior", EVirtualVoiceQueueBehaviour)
    volume = WwiseProperty[float]("Volume", float)



# Copyright 2024 Matheus Vilano
# SPDX-License-Identifier: Apache-2.0

class SingletonMeta(type):
    """Metaclass for singletons."""
    
    _instances = dict()
    """A dictionary where the keys are classes and the values are instances."""
    
    def __call__(cls, *args, **kwargs):
        """
        Operator `()`.
        :param args: Positional arguments.
        :param kwargs: Keyword arguments.
        :return: The singleton instance of the `cls` class.
        """
        if cls not in cls._instances:
            instance = super().__call__(*args, **kwargs)
            cls._instances[cls] = instance
        return cls._instances[cls]


class StaticMeta(type):
    """Metaclass for statics."""
    
    def __call__(cls, *args, **kwargs):
        """
        Operator `()`. Will prevent the creation of a new instance.
        :param args: Positional arguments.
        :param kwargs: Keyword arguments.
        :raise: TypeError, as a static class must not have any instances of itself.
        """
        raise TypeError(f"Cannot instantiate a static class. Use '{cls.__name__}.function_name()' instead.")



# Copyright 2024 Matheus Vilano
# SPDX-License-Identifier: Apache-2.0

from typing import Any as _Any

from simplevent import RefEvent as _RefEvent
from waapi import WaapiClient as _WaapiClient

from pywwise.aliases import SystemPath
from pywwise.decorators import callback, debug_build_only
from pywwise.enums import EBitDepth, ESampleRate, ESpeakerBitMask, EWaveformStr


class Debug:
    """ak.wwise.debug"""
    
    def __init__(self, client: _WaapiClient, is_debug_build: bool = False):
        """
        Constructor.
        :param client: The WAAPI client to use.
        :param is_debug_build: Should be set to true if the instance of Wwise is a debug build and debug-only
                               functions/topics are required.
        """
        self._client = client
        
        self._is_debug_build = is_debug_build
        
        self.assert_failed = _RefEvent(str, str, int, str, str)
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_wwise_debug_assertfailed.html
        \nSent when an assert has failed. **This is only available in Debug builds**.
        \n**Event Data**
        \n- The expression that failed.
        \n- The name of the source file.
        \n- The line number.
        \n- The callstack from the location of the assert.
        \n- An explanatory message accompanying the assert. May be empty.
        """
        
        if self._is_debug_build:
            self._assert_failed = self._client.subscribe("ak.wwise.debug.assertFailed", self._on_assert_failed)
    
    @callback
    def _on_assert_failed(self, event: _RefEvent, **kwargs):
        """
        Callback function for the `assertFailed` event.
        :param event: The event to broadcast.
        :param kwargs: The event data.
        """
        event(kwargs["expression"], kwargs["fileName"], int(kwargs["lineNumber"]),
              kwargs["callstack"], kwargs.get("message", ""))
    
    @debug_build_only
    def enable_asserts(self, enable: bool) -> bool:
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_wwise_debug_enableasserts.html \n
        Enables debug assertions. Every call to enableAsserts with 'false' increments the ref count. Calling
        with true decrements the ref count. **This is only available with Debug builds**.
        :param enable: Whether to enable debug assertions.
        :return: A flag indicated whether the call succeeded.
        """
        return self._client.call("ak.wwise.debug.enableAsserts", {"enable": enable}) is not None
    
    def enable_automation_mode(self, enable: bool) -> bool:
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_wwise_debug_enableautomationmode.html \n
        Enables or disables the automation mode for Wwise. This reduces the potential interruptions caused by
        message boxes and dialogs. For instance, enabling the automation mode silently accepts: project
        migration, project load log, EULA acceptance, project licence display and generic message boxes.
        :param enable: Whether to enable automation mode.
        :return: Whether the call succeeded.
        """
        return self._client.call("ak.wwise.debug.enableAutomationMode", {"enable": enable}) is not None
    
    def generate_tone_wav(self, path: SystemPath, bit_depth: EBitDepth = EBitDepth.INT_16,
                          sample_rate: ESampleRate = ESampleRate.SR_44100,
                          channel_config: ESpeakerBitMask = ESpeakerBitMask.MONO, attack_time: float = 0.0,
                          sustain_time: float = 1.0, sustain_level: float = 0.0, release_time: float = 0.0,
                          waveform: EWaveformStr = EWaveformStr.SILENCE, frequency: float = 440.0) -> bool:
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_wwise_debug_generatetonewav.html \n
        Generate a WAV file playing a tone with a simple envelope and save it to the specified location. This
        is provided as a utility to generate test WAV files.
        :param path: The path where to store the generated WAV file.
        :param bit_depth: The bit depth of the new audio file.
        :param sample_rate: The sample rate of the new audio file.
        :param channel_config: The channel configuration of the new audio file.
        :param attack_time: The attack time of the sound to generate.
        :param sustain_time: The sustain time of the sound to generate.
        :param sustain_level: The sustain level of the sound to generate. This is clamped to the range [-100.0, 0.0].
        :param release_time: The release time of the sound to generate.
        :param waveform: The waveform of the sound to generate.
        :param frequency: The frequency of the sound to generate. This is clamped to the range [1.0, 22000.0].
        :return: Whether the call succeeded and the file was created successfully.
        """
        if path.exists():
            raise ValueError("The provided path already exists.")
        args = {"path": str(path), "bitDepth": bit_depth.value, "sampleRate": sample_rate.value,
                "channelConfig": channel_config.get_name(), "attackTime": attack_time,
                "sustainTime": sustain_time, "sustainLevel": max(-100.0, min(sustain_level, 0.0)),
                "releaseTime": release_time, "waveform": waveform.value, "frequency": max(1.0, min(frequency, 22000.0))}
        self._client.call("ak.wwise.debug.generateToneWAV", args)
        return path.exists()
    
    def get_wal_tree(self) -> dict[str, dict[str, dict[str, _Any]]]:
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_wwise_debug_getwaltree.html \n
        Retrieves the WAL tree, which describes the nodes that are synchronized in the Sound Engine. Private use only.
        :return: If the call succeeded, the WAL tree; else, an empty dictionary.
        """
        results = self._client.call("ak.wwise.debug.getWalTree")
        return results.get("return", dict()) if results is not None else dict()
    
    def restart_waapi_servers(self) -> bool:
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_wwise_debug_restartwaapiservers.html \n
        Restart WAAPI servers. For internal use only.
        :return: Whether the call succeeded.
        """
        return self._client.call("ak.wwise.debug.restartWaapiServers") is not None
    
    @debug_build_only
    def test_assert(self) -> bool:
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_wwise_debug_testassert.html \n
        Private use only.
        :return: Whether the call succeeded.
        """
        return self._client.call("ak.wwise.debug.testAssert") is not None
    
    @debug_build_only
    def test_crash(self) -> bool:
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_wwise_debug_testcrash.html \n
        Private use only.
        :return: Whether the call succeeded.
        """
        return self._client.call("ak.wwise.debug.testCrash") is not None



# Copyright 2025 Matheus Vilano
# SPDX-License-Identifier: Apache-2.0

from pywwise.descriptors import WwiseProperty
from pywwise.enums import ETimeStretchMode, ETimeStretchStereoProcessing
from pywwise.objects.abc import WwiseObject


class WwiseTimeStretch(WwiseObject):
    """
    https://www.audiokinetic.com/en/library/edge/?source=SDK&id=wwiseobject_effect_wwise_time_stretch.html \n
    A class serving as an interface for getting/setting properties on Wwise objects. This type specifically targets
    the class represented by `EObjectType.WWISE_TIME_STRETCH`.
    """
    output_gain = WwiseProperty[float]("OutputGain", float)
    pitch_shift = WwiseProperty[float]("PitchShift", float)
    pitch_shift_random = WwiseProperty[float]("PitchShiftRandom", float)
    quality = WwiseProperty[float]("Quality", float)
    stretch_mode = WwiseProperty[ETimeStretchMode]("QualityEnable", ETimeStretchMode)
    stereo_processing = WwiseProperty[ETimeStretchStereoProcessing]("StereoProcessing", ETimeStretchStereoProcessing)
    time_stretch = WwiseProperty[float]("TimeStretch", float)
    time_stretch_random = WwiseProperty[float]("TimeStretchRandom", float)
    window_size = WwiseProperty[int]("WindowSize", int)



# Copyright 2024 Matheus Vilano
# SPDX-License-Identifier: Apache-2.0

from asyncio import new_event_loop as asyncio_new_event_loop, set_event_loop as asyncio_set_event_loop
from unittest import TestCase

from pywwise import new_waapi_connection


class PyWwiseTest(TestCase):
    """Base class for PyWwise test cases. Handles instantiating and deleting Ak instances automatically."""
    
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        asyncio_set_event_loop(asyncio_new_event_loop())
        cls.ak = new_waapi_connection()
    
    @classmethod
    def tearDownClass(cls):
        super().tearDownClass()
        del cls.ak



# Copyright 2025 Matheus Vilano
# SPDX-License-Identifier: Apache-2.0

from pywwise.enums import ESoundSeedAirChannelMask, ENoiseColour
from pywwise.descriptors import WwiseProperty
from pywwise.objects.abc import WwiseObject


class SoundSeedAirWoosh(WwiseObject):
    """
    https://www.audiokinetic.com/en/library/edge/?source=SDK&id=wwiseobject_source_soundseed_air_woosh.html \n
    A class serving as an interface for getting/setting properties on Wwise objects. This type specifically targets
    the class represented by `EObjectType.SOURCE_PLUGIN`.
    """
    attenuation_roll_of = WwiseProperty[float]("AttenuationRollOff", float)
    channel_mask = WwiseProperty[ESoundSeedAirChannelMask]("ChannelMask", ESoundSeedAirChannelMask)
    distance_attenuation = WwiseProperty[bool]("DistanceAttenuation", bool)
    duration = WwiseProperty[float]("Duration", float)
    duration_random = WwiseProperty[float]("DurationRandom", float)
    frequency_scale = WwiseProperty[float]("FrequencyScale", float)
    frequency_scale_automate = WwiseProperty[bool]("FrequencyScaleAutomate", bool)
    frequency_scale_random = WwiseProperty[float]("FrequencyScaleRandom", float)
    gain_offset = WwiseProperty[float]("GainOffset", float)
    gain_offset_automate = WwiseProperty[bool]("GainOffsetAutomate", bool)
    gain_offset_random = WwiseProperty[float]("GainOffsetRandom", float)
    min_distance = WwiseProperty[float]("MinDistance", float)
    noise_colour = WwiseProperty[ENoiseColour]("NoiseColour", ENoiseColour)
    oversampling = WwiseProperty[int]("Oversampling", int)
    playback_rate = WwiseProperty[float]("PlaybackRate", float)
    q_factor_scale = WwiseProperty[float]("QFactorScale", float)
    q_factor_scale_automate = WwiseProperty[bool]("QFactorScaleAutomate", bool)
    q_factor_scale_random = WwiseProperty[float]("QFactorScaleRandom", float)
    speed = WwiseProperty[float]("Speed", float)
    speed_automate = WwiseProperty[bool]("SpeedAutomate", bool)
    speed_point_speed_random = WwiseProperty[float]("SpeedPointSpeedRandom", float)
    speed_point_time_random = WwiseProperty[float]("SpeedPointTimeRandom", float)
    speed_random = WwiseProperty[float]("SpeedRandom", float)
    velocity_dynamic_range = WwiseProperty[float]("VelocityDynamicRange", float)



# Copyright 2025 Matheus Vilano
# SPDX-License-Identifier: Apache-2.0

from pywwise.descriptors import WwiseProperty
from pywwise.objects.types.acoustic_texture import AcousticTexture
from pywwise.objects.types.action import Action
from pywwise.objects.types.action_exception import ActionException
from pywwise.objects.types.actor_mixer import ActorMixer
from pywwise.objects.types.attenuation import Attenuation
from pywwise.objects.types.audio_device import AudioDevice
from pywwise.objects.types.audio_source import AudioSource
from pywwise.objects.types.aux_bus import AuxBus
from pywwise.objects.types.blend_container import BlendContainer
from pywwise.objects.types.blend_track import BlendTrack
from pywwise.objects.types.bus import Bus
from pywwise.objects.types.control_surface_binding import ControlSurfaceBinding
from pywwise.objects.types.control_surface_binding_group import ControlSurfaceBindingGroup
from pywwise.objects.types.control_surface_session import ControlSurfaceSession
from pywwise.objects.types.conversion import Conversion
from pywwise.objects.types.curve import Curve
from pywwise.objects.types.custom_state import CustomState
from pywwise.objects.types.dialogue_event import DialogueEvent
from pywwise.objects.types.effect import Effect
from pywwise.objects.types.effect_slot import EffectSlot
from pywwise.objects.types.event import Event
from pywwise.objects.types.external_source import ExternalSource
from pywwise.objects.types.external_source_file import ExternalSourceFile
from pywwise.objects.types.folder import Folder
from pywwise.objects.types.game_parameter import GameParameter
from pywwise.objects.types.language import Language
from pywwise.objects.types.marker import Marker
from pywwise.objects.types.metadata import Metadata
from pywwise.objects.types.midi_file_source import MidiFileSource
from pywwise.objects.types.midi_parameter import MidiParameter
from pywwise.objects.types.mixing_session import MixingSession
from pywwise.objects.types.modifier import Modifier
from pywwise.objects.types.modulator_envelope import ModulatorEnvelope
from pywwise.objects.types.modulator_lfo import ModulatorLfo
from pywwise.objects.types.modulator_time import ModulatorTime
from pywwise.objects.types.multi_switch_entry import MultiSwitchEntry
from pywwise.objects.types.music_clip import MusicClip
from pywwise.objects.types.music_clip_midi import MusicClipMidi
from pywwise.objects.types.music_cue import MusicCue
from pywwise.objects.types.music_event_cue import MusicEventCue
from pywwise.objects.types.music_fade import MusicFade
from pywwise.objects.types.music_playlist_container import MusicPlaylistContainer
from pywwise.objects.types.music_playlist_item import MusicPlaylistItem
from pywwise.objects.types.music_segment import MusicSegment
from pywwise.objects.types.music_stinger import MusicStinger
from pywwise.objects.types.music_switch_container import MusicSwitchContainer
from pywwise.objects.types.music_track import MusicTrack
from pywwise.objects.types.music_track_sequence import MusicTrackSequence
from pywwise.objects.types.music_transition import MusicTransition
from pywwise.objects.types.object_setting_assoc import ObjectSettingAssoc
from pywwise.objects.types.panner import Panner
from pywwise.objects.types.path_2d import Path2D
from pywwise.objects.types.platform import Platform
from pywwise.objects.types.plugin_data_source import PluginDataSource
from pywwise.objects.types.position import Position
from pywwise.objects.types.project import Project
from pywwise.objects.types.query import Query
from pywwise.objects.types.random_sequence_container import RandomSequenceContainer
from pywwise.objects.types.rtpc import Rtpc
from pywwise.objects.types.search_criteria import SearchCriteria
from pywwise.objects.types.sound import Sound
from pywwise.objects.types.sound_bank import SoundBank
from pywwise.objects.types.soundcaster_session import SoundcasterSession
from pywwise.objects.types.source_plugin import SourcePlugin
from pywwise.objects.types.state import State
from pywwise.objects.types.state_group import StateGroup
from pywwise.objects.types.switch import Switch
from pywwise.objects.types.switch_container import SwitchContainer
from pywwise.objects.types.switch_group import SwitchGroup
from pywwise.objects.types.trigger import Trigger
from pywwise.objects.types.user_project_settings import UserProjectSettings
from pywwise.objects.types.work_unit import WorkUnit


# Injections - solution for circular imports; add as needed, as a last resort. Also document in class.
MusicPlaylistContainer.transition_root = WwiseProperty[MusicTransition]("TransitionRoot", MusicTransition)
MusicStinger.segment = WwiseProperty[MusicSegment]("Segment", MusicSegment)
MusicSwitchContainer.transition_root = WwiseProperty[MusicTransition]("TransitionRoot", MusicTransition)
MusicTrack.transition_root = WwiseProperty[MusicTransition]("TransitionRoot", MusicTransition)



# Copyright 2025 Matheus Vilano
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

from typing import TYPE_CHECKING, Union as _Union

if TYPE_CHECKING:
    from pywwise.objects.types.multi_switch_entry import MultiSwitchEntry

from pywwise.descriptors import WwiseProperty
from pywwise.enums import (EColour, ECustomCueJumpMatchMode, EDestinationContextType, EMusicDestinationJumpTo,
                           EMusicDestinationSyncTo, EMusicSourceExitPoint)
from pywwise.objects.abc import WwiseObject
from pywwise.objects.types.music_playlist_container import MusicPlaylistContainer
from pywwise.objects.types.music_segment import MusicSegment
from pywwise.objects.types.music_switch_container import MusicSwitchContainer
from pywwise.objects.types.music_track import MusicTrack

_InteractiveMusicType = _Union[MusicPlaylistContainer, MusicSegment, MusicSwitchContainer, MusicTrack]

_InteractiveMusicTypeTuple = (MusicPlaylistContainer, MusicSegment, MusicSwitchContainer, MusicTrack)


class MusicTransition(WwiseObject):
    """
    https://www.audiokinetic.com/en/library/edge/?source=SDK&id=wwiseobject_musictransition.html \n
    A class serving as an interface for getting/setting properties on Wwise objects. This type specifically targets
    the class represented by `EObjectType.MUSIC_TRANSITION`.
    """
    colour = WwiseProperty[EColour]("Color", EColour)
    destination_context_object = WwiseProperty[_InteractiveMusicType]("DestinationContextObject",
                                                                      _InteractiveMusicTypeTuple)
    destination_context_type = WwiseProperty[EDestinationContextType]("DestinationContextType", EDestinationContextType)
    destination_jump_position_preset = WwiseProperty[EMusicDestinationSyncTo](
        "DestinationJumpPositionPreset", EMusicDestinationSyncTo)
    destination_playlist_jump_to = WwiseProperty[EMusicDestinationJumpTo]("DestinationPlaylistJumpTo",
                                                                          EMusicDestinationJumpTo)
    enable_destination_fade_in = WwiseProperty[bool]("EnableDestinationFadeIn", bool)
    enable_source_fade_out = WwiseProperty[bool]("EnableSourceFadeOut", bool)
    enable_transition_fade_in = WwiseProperty[bool]("EnableTransitionFadeIn", bool)
    enable_transition_fade_out = WwiseProperty[bool]("EnableTransitionFadeOut", bool)
    entries: WwiseProperty[tuple[MultiSwitchEntry, ...]] = WwiseProperty(
        "Entries", tuple)  # Using `:` to avoid circular imports.
    exit_source_at = WwiseProperty[EMusicSourceExitPoint]("ExitSourceAt", EMusicSourceExitPoint)
    exit_source_custom_cue_match_name = WwiseProperty[str]("ExitSourceCustomCueMatchName", str)
    is_folder = WwiseProperty[bool]("IsFolder", bool)
    jump_to_custom_cue_match_mode = WwiseProperty[ECustomCueJumpMatchMode](
        "JumpToCustomCueMatchMode", ECustomCueJumpMatchMode)
    jump_to_custom_cue_match_name = WwiseProperty[str]("jumpToCustomCueMatchName", str)
    override_colour = WwiseProperty[bool]("OverrideColor", bool)
    play_destination_pre_entry = WwiseProperty[bool]("PlayDestinationPreEntry", bool)
    play_post_exit = WwiseProperty[bool]("PlayPostExit", bool)
    play_transition_post_Exit = WwiseProperty[bool]("PlayTransitionPostExit", bool)
    play_transition_pre_entry = WwiseProperty[bool]("PlayTransitionPreEntry", bool)
    source_context_object = WwiseProperty[_InteractiveMusicType]("SourceContextObject", _InteractiveMusicTypeTuple)
    source_context_type = WwiseProperty[EDestinationContextType]("SourceContextType", EDestinationContextType)
    use_transition_object = WwiseProperty[bool]("UseTransitionObject", bool)



# Copyright 2024 Matheus Vilano
# SPDX-License-Identifier: Apache-2.0

from typing import Callable as _Callable


def callback(func: _Callable) -> _Callable:
    """
    A decorator for PyWwise callbacks. Checks if the associated event is valid and has any subscribers. If it has at
    least one subscriber, it will be broadcast. For the decorated function to work properly: \n
    - its name should start with the prefix `_on_` (e.g. `_on_event_happened`).
    - its associated event should have the same name, minus the prefix `_on_` (e.g. `event_happened`).
    - both the decorated function and the associated event should be encapsulated within the same object.
    :param func: The function to decorate.
    :return: The decorated function.
    """
    
    def wrapper(self, **kwargs):
        event = getattr(self, func.__name__[4:])  # removing the prefix "_on_" gives the RefEvent
        if event is None:
            raise NameError(f"Function '{func.__name__}' missing associated event: '{func.__name__[4:]}'.")
        if len(event) > 0:
            return func(self, event, **kwargs)
    
    return wrapper


def console_instance_only(func: _Callable) -> _Callable:
    """
    A decorator for functions that can only be called on a console instance of Wwise. Checks the owner object's
    `is_console_instance` (or, alternatively, `_is_console_instance`); if `None` or `False`, the decorated function will
    NOT be called and an exception is raised. Note: if there is no owner object, the check defaults to `None`. \n
    :param func: The function to decorate.
    :raise: NameError - if `is_console_instance` does not exist.
    :raise: ValueError - if `is_console_instance` is False.
    :return: The decorated function.
    """
    attr_name = "is_console_instance"
    
    def wrapper(self, *args, **kwargs):
        flag = getattr(self, attr_name) \
            if hasattr(self, attr_name) \
            else getattr(self, f"_{attr_name}", None)
        if flag is None:
            raise NameError(f"Object '{self}' does not have an attribute named '{attr_name}'.")
        if flag is True:
            raise ValueError(f"Debug function called on a non-console instance of 'Ak'.")
        return func(self, *args, **kwargs)
    
    return wrapper


def debug_build_only(func: _Callable) -> _Callable:
    """
    A decorator for PyWwise debug-only functions. Checks the owner object's `is_debug_build` (or, alternatively,
    `_is_debug_build`); if `None` or `False`, the decorated function will NOT be called and an exception is raised.
    Note: if there is no owner object, the check defaults to `None`. \n
    :param func: The function to decorate.
    :raise: NameError - if `is_debug_build` does not exist.
    :raise: ValueError - if `is_debug_build` is False.
    :return: The decorated function.
    """
    attr_name = "is_debug_build"
    
    def wrapper(self, *args, **kwargs):
        flag = getattr(self, attr_name) \
            if hasattr(self, attr_name) \
            else getattr(self, f"_{attr_name}", None)
        if flag is None:
            raise NameError(f"Object '{self}' does not have an attribute named '{attr_name}'.")
        if flag is True:
            raise ValueError(f"Debug function called on a non-debug instance of 'Ak'.")
        return func(self, *args, **kwargs)
    
    return wrapper



# Copyright 2024 Matheus Vilano
# SPDX-License-Identifier: Apache-2.0

from waapi import WaapiClient as _WaapiClient

from pywwise.aliases import ListOrTuple, SystemPath
from pywwise.decorators import console_instance_only
from pywwise.enums import EBasePlatform
from pywwise.primitives import Name
from pywwise.structs import PlatformInfo


class Project:
    """ak.wwise.console.project"""
    
    def __init__(self, client: _WaapiClient, is_console_instance: bool = False):
        """
        Constructor.
        :param client: The WAAPI client to use.
        :param is_console_instance: Should be set to true if the instance of Wwise is running in a console window.
        """
        self._client = client
        self._is_console_instance = is_console_instance
    
    @console_instance_only
    def close(self) -> bool:
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_wwise_console_project_close.html \n
        Closes the current project. This operation is synchronous.
        :return: True if there was a project open and it closed successfully; otherwise, false.
        """
        results = self._client.call("ak.wwise.console.project.close")
        return results.get("hadProjectOpen", False) if results is not None else False
    
    @console_instance_only
    def create(self, project_path: SystemPath,
               platforms: ListOrTuple[PlatformInfo] = (PlatformInfo("Windows", EBasePlatform.WINDOWS),),
               languages: ListOrTuple[Name] = tuple(Name("English(US)"))) -> bool:
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_wwise_console_project_create.html \n
        Creates, saves and opens new empty project, specified by path and platform. The project has no
        factory setting WorkUnit. This operation is synchronous.
        :param project_path: The path to the project WPROJ file. Example: "C:/Projects/MyProject/MyProject.wproj".
        :param platforms: Specifies the platform(s) supported by the new project. If not specified, only Windows is
                          used.
        :param languages: Array of languages to creates for this project. If not specified, English(US) becomes the
                          default language. When multiple languages are specified, the first one becomes the default
                          language.
        :return: Whether the call succeeded. True does not necessarily mean the project was successfully created.
        """
        platforms = list(dict.fromkeys(platforms))
        args = {"path": str(project_path), "platforms": list(), "languages": [language for language in languages]}
        for platform in platforms:
            args["platforms"].append({"name": platform.name, "basePlatform": platform.base})
        return self._client.call("ak.wwise.console.project.create", args) is not None
    
    @console_instance_only
    def open(self, project_path: SystemPath,
             is_migration_allowed: bool,
             version_control_auto_checkout: bool = True) -> bool:
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_wwise_console_project_open.html \n
        Opens a project, specified by path. This operation is synchronous.
        :param project_path: The path to the project WPROJ file.
        :param is_migration_allowed: Whether migration is allowed.
        :param version_control_auto_checkout: Determines if Wwise automatically performs a Checkout source control
                                              operation for affected work units and for the project. Only supported in
                                              Wwise 2023 or above.
        :return: Whether the call succeeded. True does not necessarily mean the project was successfully opened.
        """
        migration_action = "migrate" if is_migration_allowed else "fail"
        args = {"path": str(project_path),
                "onMigrationRequired": migration_action,
                **({"autoCheckOutToSourceControl": False} if not version_control_auto_checkout else {})}
        return self._client.call("ak.wwise.console.project.open", args) is not None



# Copyright 2025 Matheus Vilano
# SPDX-License-Identifier: Apache-2.0

from pywwise.enums import ESoundSeedAirChannelMask
from pywwise.descriptors import WwiseProperty
from pywwise.objects.abc import WwiseObject


class SoundSeedAirWind(WwiseObject):
    """
    https://www.audiokinetic.com/en/library/edge/?source=SDK&id=wwiseobject_source_soundseed_air_wind.html \n
    A class serving as an interface for getting/setting properties on Wwise objects. This type specifically targets
    the class represented by `EObjectType.SOURCE_PLUGIN`.
    """
    attenuation_roll_of = WwiseProperty[float]("AttenuationRollOff", float)
    average_velocity = WwiseProperty[float]("AverageVelocity", float)
    average_velocity_automate = WwiseProperty[bool]("AverageVelocityAutomate", bool)
    average_velocity_random = WwiseProperty[float]("AverageVelocityRandom", float)
    channel_mask = WwiseProperty[ESoundSeedAirChannelMask]("ChannelMask", ESoundSeedAirChannelMask)
    direction = WwiseProperty[float]("Direction", float)
    direction_automate = WwiseProperty[bool]("DirectionAutomate", bool)
    direction_random = WwiseProperty[float]("DirectionRandom", float)
    duration = WwiseProperty[float]("Duration", float)
    duration_random = WwiseProperty[float]("DurationRandom", float)
    frequency_scale = WwiseProperty[float]("FrequencyScale", float)
    frequency_scale_automate = WwiseProperty[bool]("FrequencyScaleAutomate", bool)
    frequency_scale_random = WwiseProperty[float]("FrequencyScaleRandom", float)
    gain_offset = WwiseProperty[float]("GainOffset", float)
    gain_offset_automate = WwiseProperty[bool]("GainOffsetAutomate", bool)
    gain_offset_random = WwiseProperty[float]("GainOffsetRandom", float)
    gustiness = WwiseProperty[float]("Gustiness", float)
    gustiness_automate = WwiseProperty[bool]("GustinessAutomate", bool)
    gustiness_random = WwiseProperty[float]("GustinessRandom", float)
    max_distance = WwiseProperty[float]("MaxDistance", float)
    min_distance = WwiseProperty[float]("MinDistance", float)
    playback_rate = WwiseProperty[float]("PlaybackRate", float)
    q_factor_scale = WwiseProperty[float]("QFactorScale", float)
    q_factor_scale_automate = WwiseProperty[bool]("QFactorScaleAutomate", bool)
    q_factor_scale_random = WwiseProperty[float]("QFactorScaleRandom", float)
    variability = WwiseProperty[float]("Variability", float)
    variability_automate = WwiseProperty[bool]("VariabilityAutomate", bool)
    variability_random = WwiseProperty[float]("VariabilityRandom", float)
    velocity_dynamic_range = WwiseProperty[float]("VelocityDynamicRange", float)



# Copyright 2025 Matheus Vilano
# SPDX-License-Identifier: Apache-2.0

from pywwise.descriptors import WwiseProperty
from pywwise.enums import (ESoundSeedGrainChannelConfig, ESoundSeedGrainDurationLink, ESoundSeedGrainEnvelopeType,
                           ESoundSeedGrainFilterType, ESoundSeedGrainFrequencyTime, ESoundSeedGrainPositioning,
                           ESoundSeedGrainQuantization, ESoundSeedGrainQuantizationSimple, ESoundSeedGrainSelection,
                           ESoundSeedGrainWaveform, ESoundSeedGrainWindowMode)
from pywwise.objects.abc import WwiseObject


class SoundSeedGrain(WwiseObject):
    """
    https://www.audiokinetic.com/en/library/edge/?source=SDK&id=wwiseobject_source_soundseed_grain.html \n
    A class serving as an interface for getting/setting properties on Wwise objects. This type specifically targets
    the class represented by `EObjectType.SOURCE_PLUGIN`.
    """
    amplitude = WwiseProperty[float]("Amplitude", float)
    amplitude_mod1_depth = WwiseProperty[float]("AmplitudeMod1Depth", float)
    amplitude_mod1_quantization = WwiseProperty[ESoundSeedGrainQuantizationSimple](
        "AmplitudeMod1Quantization", ESoundSeedGrainQuantizationSimple)
    amplitude_mod2_depth = WwiseProperty[float]("AmplitudeMod2Depth", float)
    amplitude_mod2_quantization = WwiseProperty[ESoundSeedGrainQuantizationSimple](
        "AmplitudeMod2Quantization", ESoundSeedGrainQuantizationSimple)
    amplitude_mod3_depth = WwiseProperty[float]("AmplitudeMod3Depth", float)
    amplitude_mod3_quantization = WwiseProperty[ESoundSeedGrainQuantizationSimple](
        "AmplitudeMod3Quantization", ESoundSeedGrainQuantizationSimple)
    amplitude_mod4_depth = WwiseProperty[float]("AmplitudeMod4Depth", float)
    amplitude_mod4_quantization = WwiseProperty[ESoundSeedGrainQuantizationSimple](
        "AmplitudeMod4Quantization", ESoundSeedGrainQuantizationSimple)
    attack = WwiseProperty[float]("Attack", float)
    attack_mod1_depth = WwiseProperty[float]("AttackMod1Depth", float)
    attack_mod1_quantization = WwiseProperty[ESoundSeedGrainQuantization](
        "AttackMod1Quantization", ESoundSeedGrainQuantization)
    attack_mod2_depth = WwiseProperty[float]("AttackMod2Depth", float)
    attack_mod2_quantization = WwiseProperty[ESoundSeedGrainQuantization](
        "AttackMod2Quantization", ESoundSeedGrainQuantization)
    attack_mod3_depth = WwiseProperty[float]("AttackMod13Depth", float)
    attack_mod3_quantization = WwiseProperty[ESoundSeedGrainQuantization](
        "AttackMod3Quantization", ESoundSeedGrainQuantization)
    attack_mod4_depth = WwiseProperty[float]("AttackMod4Depth", float)
    attack_mod4_quantization = WwiseProperty[ESoundSeedGrainQuantization](
        "AttackMod4Quantization", ESoundSeedGrainQuantization)
    azimuth = WwiseProperty[float]("Azimuth", float)
    azimuth_mod1_depth = WwiseProperty[float]("AzimuthMod1Depth", float)
    azimuth_mod1_quantization = WwiseProperty[ESoundSeedGrainQuantizationSimple](
        "AzimuthMod1Quantization", ESoundSeedGrainQuantizationSimple)
    azimuth_mod2_depth = WwiseProperty[float]("AzimuthMod2Depth", float)
    azimuth_mod2_quantization = WwiseProperty[ESoundSeedGrainQuantizationSimple](
        "AzimuthMod2Quantization", ESoundSeedGrainQuantizationSimple)
    azimuth_mod3_depth = WwiseProperty[float]("AzimuthMod3Depth", float)
    azimuth_mod3_quantization = WwiseProperty[ESoundSeedGrainQuantizationSimple](
        "AzimuthMod3Quantization", ESoundSeedGrainQuantizationSimple)
    azimuth_mod4_depth = WwiseProperty[float]("AzimuthMod4Depth", float)
    azimuth_mod4_quantization = WwiseProperty[ESoundSeedGrainQuantizationSimple](
        "AzimuthMod4Quantization", ESoundSeedGrainQuantizationSimple)
    duration = WwiseProperty[float]("Duration", float)
    duration_link = WwiseProperty[ESoundSeedGrainDurationLink]("DurationLink", ESoundSeedGrainDurationLink)
    duration_mod1_depth = WwiseProperty[float]("DurationMod1Depth", float)
    duration_mod1_quantization = WwiseProperty[ESoundSeedGrainQuantization](
        "DurationMod1Quantization", ESoundSeedGrainQuantization)
    duration_mod2_depth = WwiseProperty[float]("DurationMod2Depth", float)
    duration_mod2_quantization = WwiseProperty[ESoundSeedGrainQuantization](
        "DurationMod2Quantization", ESoundSeedGrainQuantization)
    duration_mod3_depth = WwiseProperty[float]("DurationMod3Depth", float)
    duration_mod3_quantization = WwiseProperty[ESoundSeedGrainQuantization](
        "DurationMod3Quantization", ESoundSeedGrainQuantization)
    duration_mod4_depth = WwiseProperty[float]("DurationMod4Depth", float)
    duration_mod4_quantization = WwiseProperty[ESoundSeedGrainQuantization](
        "DurationMod4Quantization", ESoundSeedGrainQuantization)
    duration_multiplier = WwiseProperty[float]("DurationMultiplier", float)
    duration_multiplier_mod1_depth = WwiseProperty[float]("DurationMultiplierMod1Depth", float)
    duration_multiplier_mod1_quantization = WwiseProperty[ESoundSeedGrainQuantizationSimple](
        "DurationMultiplierMod1Quantization", ESoundSeedGrainQuantizationSimple)
    duration_multiplier_mod2_depth = WwiseProperty[float]("DurationMultiplierMod2Depth", float)
    duration_multiplier_mod2_quantization = WwiseProperty[ESoundSeedGrainQuantizationSimple](
        "DurationMultiplierMod2Quantization", ESoundSeedGrainQuantizationSimple)
    duration_multiplier_mod3_depth = WwiseProperty[float]("DurationMultiplierMod3Depth", float)
    duration_multiplier_mod3_quantization = WwiseProperty[ESoundSeedGrainQuantizationSimple](
        "DurationMultiplierMod3Quantization", ESoundSeedGrainQuantizationSimple)
    duration_multiplier_mod4_depth = WwiseProperty[float]("DurationMultiplierMod4Depth", float)
    duration_multiplier_mod4_quantization = WwiseProperty[ESoundSeedGrainQuantizationSimple](
        "DurationMultiplierMod4Quantization", ESoundSeedGrainQuantizationSimple)
    elevation = WwiseProperty[float]("Elevation", float)
    elevation_mod1_depth = WwiseProperty[float]("ElevationMod1Depth", float)
    elevation_mod1_quantization = WwiseProperty[ESoundSeedGrainQuantizationSimple](
        "ElevationMod1Quantization", ESoundSeedGrainQuantizationSimple)
    elevation_mod2_depth = WwiseProperty[float]("ElevationMod2Depth", float)
    elevation_mod2_quantization = WwiseProperty[ESoundSeedGrainQuantizationSimple](
        "ElevationMod2Quantization", ESoundSeedGrainQuantizationSimple)
    elevation_mod3_depth = WwiseProperty[float]("ElevationMod3Depth", float)
    elevation_mod3_quantization = WwiseProperty[ESoundSeedGrainQuantizationSimple](
        "ElevationMod3Quantization", ESoundSeedGrainQuantizationSimple)
    elevation_mod4_depth = WwiseProperty[float]("ElevationMod4Depth", float)
    elevation_mod4_quantization = WwiseProperty[ESoundSeedGrainQuantizationSimple](
        "ElevationMod4Quantization", ESoundSeedGrainQuantizationSimple)
    envelope_type = WwiseProperty[ESoundSeedGrainEnvelopeType]("EnvelopeType", ESoundSeedGrainEnvelopeType)
    filter_frequency = WwiseProperty[float]("FilterFreq", float)
    filter_frequency_mod1_depth = WwiseProperty[float]("FilterFreqMod1Depth", float)
    filter_frequency_mod1_quantization = WwiseProperty[ESoundSeedGrainQuantization](
        "FilterFreqMod1Quantization", ESoundSeedGrainQuantization)
    filter_frequency_mod2_depth = WwiseProperty[float]("FilterFreqMod2Depth", float)
    filter_frequency_mod2_quantization = WwiseProperty[ESoundSeedGrainQuantization](
        "FilterFreqMod2Quantization", ESoundSeedGrainQuantization)
    filter_frequency_mod3_depth = WwiseProperty[float]("FilterFreqMod3Depth", float)
    filter_frequency_mod3_quantization = WwiseProperty[ESoundSeedGrainQuantization](
        "FilterFreqMod3Quantization", ESoundSeedGrainQuantization)
    filter_frequency_mod4_depth = WwiseProperty[float]("FilterFreqMod4Depth", float)
    filter_frequency_mod4_quantization = WwiseProperty[ESoundSeedGrainQuantization](
        "FilterFreqMod4Quantization", ESoundSeedGrainQuantization)
    filter_q = WwiseProperty[float]("FilterQ", float)
    filter_q_mod1_depth = WwiseProperty[float]("FilterQMod1Depth", float)
    filter_q_mod1_quantization = WwiseProperty[ESoundSeedGrainQuantizationSimple](
        "FilterQMod1Quantization", ESoundSeedGrainQuantizationSimple)
    filter_q_mod2_depth = WwiseProperty[float]("FilterQMod2Depth", float)
    filter_q_mod2_quantization = WwiseProperty[ESoundSeedGrainQuantizationSimple](
        "FilterQMod2Quantization", ESoundSeedGrainQuantizationSimple)
    filter_q_mod3_depth = WwiseProperty[float]("FilterQMod3Depth", float)
    filter_q_mod3_quantization = WwiseProperty[ESoundSeedGrainQuantizationSimple](
        "FilterQMod3Quantization", ESoundSeedGrainQuantizationSimple)
    filter_q_mod4_depth = WwiseProperty[float]("FilterQMod4Depth", float)
    filter_q_mod4_quantization = WwiseProperty[ESoundSeedGrainQuantizationSimple](
        "FilterQMod4Quantization", ESoundSeedGrainQuantizationSimple)
    filter_type = WwiseProperty[ESoundSeedGrainFilterType]("FilterType", ESoundSeedGrainFilterType)
    grain_rate = WwiseProperty[float]("GrainRate", float)
    grain_rate_mod1_depth = WwiseProperty[float]("GrainRateMod1Depth", float)
    grain_rate_mod1_quantization = WwiseProperty[ESoundSeedGrainQuantization](
        "GrainRateMod1Quantization", ESoundSeedGrainQuantization)
    grain_rate_mod2_depth = WwiseProperty[float]("GrainRateMod2Depth", float)
    grain_rate_mod2_quantization = WwiseProperty[ESoundSeedGrainQuantization](
        "GrainRateMod2Quantization", ESoundSeedGrainQuantization)
    grain_rate_mod3_depth = WwiseProperty[float]("GrainRateMod3Depth", float)
    grain_rate_mod3_quantization = WwiseProperty[ESoundSeedGrainQuantization](
        "GrainRateMod3Quantization", ESoundSeedGrainQuantization)
    grain_rate_mod4_depth = WwiseProperty[float]("GrainRateMod4Depth", float)
    grain_rate_mod4_quantization = WwiseProperty[ESoundSeedGrainQuantization](
        "GrainRateMod4Quantization", ESoundSeedGrainQuantization)
    grain_time = WwiseProperty[float]("GrainTime", float)
    grain_time_mod1_depth = WwiseProperty[float]("GrainTimeMod1Depth", float)
    grain_time_mod1_quantization = WwiseProperty[ESoundSeedGrainQuantization](
        "GrainTimeMod1Quantization", ESoundSeedGrainQuantization)
    grain_time_mod2_depth = WwiseProperty[float]("GrainTimeMod2Depth", float)
    grain_time_mod2_quantization = WwiseProperty[ESoundSeedGrainQuantization](
        "GrainTimeMod2Quantization", ESoundSeedGrainQuantization)
    grain_time_mod3_depth = WwiseProperty[float]("GrainTimeMod3Depth", float)
    grain_time_mod3_quantization = WwiseProperty[ESoundSeedGrainQuantization](
        "GrainTimeMod3Quantization", ESoundSeedGrainQuantization)
    grain_time_mod4_depth = WwiseProperty[float]("GrainTimeMod4Depth", float)
    grain_time_mod4_quantization = WwiseProperty[ESoundSeedGrainQuantization](
        "GrainTimeMod4Quantization", ESoundSeedGrainQuantization)
    marker_select = WwiseProperty[float]("MarkerSelect", float)
    marker_select_mod1_depth = WwiseProperty[float]("MarkerSelectMod1Depth", float)
    marker_select_mod1_quantization = WwiseProperty[ESoundSeedGrainQuantizationSimple](
        "MarkerSelectMod1Quantization", ESoundSeedGrainQuantizationSimple)
    marker_select_mod2_depth = WwiseProperty[float]("MarkerSelectMod2Depth", float)
    marker_select_mod2_quantization = WwiseProperty[ESoundSeedGrainQuantizationSimple](
        "MarkerSelectMod2Quantization", ESoundSeedGrainQuantizationSimple)
    marker_select_mod3_depth = WwiseProperty[float]("MarkerSelectMod3Depth", float)
    marker_select_mod3_quantization = WwiseProperty[ESoundSeedGrainQuantizationSimple](
        "MarkerSelectMod3Quantization", ESoundSeedGrainQuantizationSimple)
    marker_select_mod4_depth = WwiseProperty[float]("MarkerSelectMod4Depth", float)
    marker_select_mod4_quantization = WwiseProperty[ESoundSeedGrainQuantizationSimple](
        "MarkerSelectMod4Quantization", ESoundSeedGrainQuantizationSimple)
    max_num_grains = WwiseProperty[int]("MaxNumGrains", int)
    midi_map_transpose = WwiseProperty[bool]("MidiMapTranspose", bool)
    mod_amount1 = WwiseProperty[float]("ModAmount1", float)
    mod_amount2 = WwiseProperty[float]("ModAmount2", float)
    mod_amount3 = WwiseProperty[float]("ModAmount3", float)
    mod_amount4 = WwiseProperty[float]("ModAmount4", float)
    mod_period1 = WwiseProperty[float]("ModPeriod1", float)
    mod_period2 = WwiseProperty[float]("ModPeriod2", float)
    mod_period3 = WwiseProperty[float]("ModPeriod3", float)
    mod_period4 = WwiseProperty[float]("ModPeriod4", float)
    mod_rate1 = WwiseProperty[float]("ModRate1", float)
    mod_rate2 = WwiseProperty[float]("ModRate2", float)
    mod_rate3 = WwiseProperty[float]("ModRate3", float)
    mod_rate4 = WwiseProperty[float]("ModRate4", float)
    mod_select1 = WwiseProperty[ESoundSeedGrainSelection]("ModSelect1", ESoundSeedGrainSelection)
    mod_select2 = WwiseProperty[ESoundSeedGrainSelection]("ModSelect2", ESoundSeedGrainSelection)
    mod_select3 = WwiseProperty[ESoundSeedGrainSelection]("ModSelect3", ESoundSeedGrainSelection)
    mod_select4 = WwiseProperty[ESoundSeedGrainSelection]("ModSelect4", ESoundSeedGrainSelection)
    mod_waveform1 = WwiseProperty[ESoundSeedGrainWaveform]("ModWaveform1", ESoundSeedGrainWaveform)
    mod_waveform2 = WwiseProperty[ESoundSeedGrainWaveform]("ModWaveform2", ESoundSeedGrainWaveform)
    mod_waveform3 = WwiseProperty[ESoundSeedGrainWaveform]("ModWaveform3", ESoundSeedGrainWaveform)
    mod_waveform4 = WwiseProperty[ESoundSeedGrainWaveform]("ModWaveform4", ESoundSeedGrainWaveform)
    offset = WwiseProperty[float]("Offset", float)
    offset_mod1_depth = WwiseProperty[float]("OffsetMod1Depth", float)
    offset_mod1_quantization = WwiseProperty[ESoundSeedGrainQuantizationSimple](
        "OffsetMod1Quantization", ESoundSeedGrainQuantizationSimple)
    offset_mod2_depth = WwiseProperty[float]("OffsetMod2Depth", float)
    offset_mod2_quantization = WwiseProperty[ESoundSeedGrainQuantizationSimple](
        "OffsetMod2Quantization", ESoundSeedGrainQuantizationSimple)
    offset_mod3_depth = WwiseProperty[float]("OffsetMod3Depth", float)
    offset_mod3_quantization = WwiseProperty[ESoundSeedGrainQuantizationSimple](
        "OffsetMod3Quantization", ESoundSeedGrainQuantizationSimple)
    offset_mod4_depth = WwiseProperty[float]("OffsetMod4Depth", float)
    offset_mod4_quantization = WwiseProperty[ESoundSeedGrainQuantizationSimple](
        "OffsetMod4Quantization", ESoundSeedGrainQuantizationSimple)
    output_channel_config = WwiseProperty[ESoundSeedGrainChannelConfig](
        "OutputChannelConfig", ESoundSeedGrainChannelConfig)
    output_level = WwiseProperty[float]("OutputLevel", float)
    positioning_select = WwiseProperty[ESoundSeedGrainPositioning]("PositioningSelect", ESoundSeedGrainPositioning)
    quantize_to_markers = WwiseProperty[bool]("QuantizeToMarkers", bool)
    release = WwiseProperty[float]("Release", float)
    release_mod1_depth = WwiseProperty[float]("ReleaseMod1Depth", float)
    release_mod1_quantization = WwiseProperty[ESoundSeedGrainQuantization](
        "ReleaseMod1Quantization", ESoundSeedGrainQuantization)
    release_mod2_depth = WwiseProperty[float]("ReleaseMod2Depth", float)
    release_mod2_quantization = WwiseProperty[ESoundSeedGrainQuantization](
        "ReleaseMod2Quantization", ESoundSeedGrainQuantization)
    release_mod3_depth = WwiseProperty[float]("ReleaseMod3Depth", float)
    release_mod3_quantization = WwiseProperty[ESoundSeedGrainQuantization](
        "ReleaseMod3Quantization", ESoundSeedGrainQuantization)
    release_mod4_depth = WwiseProperty[float]("ReleaseMod4Depth", float)
    release_mod4_quantization = WwiseProperty[ESoundSeedGrainQuantization](
        "ReleaseMod4Quantization", ESoundSeedGrainQuantization)
    select_frequency_time_grain = WwiseProperty[ESoundSeedGrainFrequencyTime](
        "SelectFreqTimeGrain", ESoundSeedGrainFrequencyTime)
    speed = WwiseProperty[float]("Speed", float)
    speed_mod1_depth = WwiseProperty[float]("SpeedMod1Depth", float)
    speed_mod1_quantization = WwiseProperty[ESoundSeedGrainQuantizationSimple](
        "SpeedMod1Quantization", ESoundSeedGrainQuantizationSimple)
    speed_mod2_depth = WwiseProperty[float]("SpeedMod2Depth", float)
    speed_mod2_quantization = WwiseProperty[ESoundSeedGrainQuantizationSimple](
        "SpeedMod2Quantization", ESoundSeedGrainQuantizationSimple)
    speed_mod3_depth = WwiseProperty[float]("SpeedMod3Depth", float)
    speed_mod3_quantization = WwiseProperty[ESoundSeedGrainQuantizationSimple](
        "SpeedMod3Quantization", ESoundSeedGrainQuantizationSimple)
    speed_mod4_depth = WwiseProperty[float]("SpeedMod4Depth", float)
    speed_mod4_quantization = WwiseProperty[ESoundSeedGrainQuantizationSimple](
        "SpeedMod4Quantization", ESoundSeedGrainQuantizationSimple)
    spread = WwiseProperty[float]("Spread", float)
    spread_mod1_depth = WwiseProperty[float]("SpreadMod1Depth", float)
    spread_mod1_quantization = WwiseProperty[ESoundSeedGrainQuantizationSimple](
        "SpreadMod1Quantization", ESoundSeedGrainQuantizationSimple)
    spread_mod2_depth = WwiseProperty[float]("SpreadMod2Depth", float)
    spread_mod2_quantization = WwiseProperty[ESoundSeedGrainQuantizationSimple](
        "SpreadMod2Quantization", ESoundSeedGrainQuantizationSimple)
    spread_mod3_depth = WwiseProperty[float]("SpreadMod3Depth", float)
    spread_mod3_quantization = WwiseProperty[ESoundSeedGrainQuantizationSimple](
        "SpreadMod3Quantization", ESoundSeedGrainQuantizationSimple)
    spread_mod4_depth = WwiseProperty[float]("SpreadMod4Depth", float)
    spread_mod4_quantization = WwiseProperty[ESoundSeedGrainQuantizationSimple](
        "SpreadMod4Quantization", ESoundSeedGrainQuantizationSimple)
    transpose = WwiseProperty[float]("Transpose", float)
    transpose_mod1_depth = WwiseProperty[float]("TransposeMod1Depth", float)
    transpose_mod1_quantization = WwiseProperty[ESoundSeedGrainQuantization](
        "TransposeMod1Quantization", ESoundSeedGrainQuantization)
    transpose_mod2_depth = WwiseProperty[float]("TransposeMod2Depth", float)
    transpose_mod2_quantization = WwiseProperty[ESoundSeedGrainQuantization](
        "TransposeMod2Quantization", ESoundSeedGrainQuantization)
    transpose_mod3_depth = WwiseProperty[float]("TransposeMod3Depth", float)
    transpose_mod3_quantization = WwiseProperty[ESoundSeedGrainQuantization](
        "TransposeMod3Quantization", ESoundSeedGrainQuantization)
    transpose_mod4_depth = WwiseProperty[float]("TransposeMod4Depth", float)
    transpose_mod4_quantization = WwiseProperty[ESoundSeedGrainQuantization](
        "TransposeMod4Quantization", ESoundSeedGrainQuantization)
    transpose_root = WwiseProperty[int]("TransposeRoot", int)
    window_mode = WwiseProperty[ESoundSeedGrainWindowMode]("WindowMode", ESoundSeedGrainWindowMode)



# Copyright 2025 Matheus Vilano
# SPDX-License-Identifier: Apache-2.0

from pywwise.descriptors import WwiseProperty
from pywwise.objects.abc import WwiseObject
from pywwise.objects.types.rtpc import Rtpc


class BlendTrack(WwiseObject):
    """
    https://www.audiokinetic.com/en/library/edge/?source=SDK&id=wwiseobject_blendtrack.html \n
    A class serving as an interface for getting/setting properties on Wwise objects. This type specifically targets
    the class represented by `EObjectType.BLEND_TRACK`.
    """
    enable_cross_fading = WwiseProperty[bool]("EnableCrossFading", bool)
    highpass = WwiseProperty[int]("Highpass", int)
    layer_cross_fade_control_input = WwiseProperty[Rtpc]("LayerCrossFadeControlInput", Rtpc)
    lowpass = WwiseProperty[int]("Lowpass", int)
    make_up_gain = WwiseProperty[float]("MakeUpGain", float)
    pitch = WwiseProperty[int]("Pitch", int)
    rtpc = WwiseProperty[tuple[Rtpc, ...]]("RTPC", tuple)
    volume = WwiseProperty[float]("Volume", float)



# Copyright 2024 Matheus Vilano
# SPDX-License-Identifier: Apache-2.0

from typing import Iterator as _Iterator, Self as _Self

from pywwise.aliases import RegexPattern
from pywwise.enums import EObjectType, EWaqlLogicalOperator, EWaqlSelectExpression
from pywwise.primitives import GUID, Name, ProjectPath, ShortID


class WaqlQuery:
    """
    A helper class for building WAQL queries. Queries may be built without this class, which is actually preferred;
    but if you are unfamiliar with WAQL, this class can be a time-saver. Useful features:\n
    - To print the query, simply pass the query to the `print` function (or similar). Example: `print(my_waql_object)`.
    - To get the generated query, convert the instance to a `str`. Example: `str(my_waql_object)`.
    - To get and set a specific query component, you can use the regular `[]` (subscription) operator, like with lists,
    tuples, etc. Example: `my_waql_object[0]`.
    - To delete a specific query component, you can use the regular `del` (delete) operator. Example: `del my_waql_obj`.
    If you are interested in how exactly this class works: it encapsulates a list of strings (`list[str]`) and most
    methods are just sugar syntax for appending `str` objects (most of which are WAQL "components") to the list. There
    are also overrides for "magic methods" to allow for some of the syntax the features described in this docstring.
    """
    
    def __init__(self):
        """Constructor for WAQL queries. To build a query, instantiate a WAQL object, then use the objects methods."""
        self._components: list[str] = []
    
    def __str__(self) -> str:
        """:return: The query (all components), formatted as a string."""
        return f"$ {' '.join(self._components)}"
    
    def __repr__(self) -> str:
        """:return: The query (all components), formatted as a string."""
        return self.__str__()
    
    def __getitem__(self, index: int) -> str:
        """
        Manually gets a component at the specified index.
        :param index: The index where to get a component.
        """
        return self._components[index]
    
    def __setitem__(self, index: int, component: str):
        """
        Manually sets a component at the specified index.
        :param index: The index at which to set a component.
        :param component: The component to set.
        """
        self._components[index] = component
    
    def __iter__(self) -> _Iterator[str]:
        """
        Generates an iterator. Useful when going over the components manually.
        :return: The generated iterator.
        """
        return iter(self._components)
    
    def __delitem__(self, index: int):
        """
        Deletes a component located at the specified index. This will actually pop the component, so the list WILL
        shift.
        :param index: Index to delete.
        """
        self._components.pop(index)
    
    def __add__(self, component: str) -> _Self:
        """
        Adds a new component to the query.
        :param component: The component to add.
        """
        self._components.append(component)
        return self
    
    def from_object(self, *objects: tuple[EObjectType, Name | ShortID] | ProjectPath | GUID):
        """
        Generates a sequence of objects from the specified objects.
        :param objects: Wwise objects; more specifically, their names, shortIDs, project paths, or GUIDs.
        """
        component = "from object "
        for obj in objects:
            if isinstance(obj, tuple):
                obj = f"{obj[0].get_type_name()}:{obj[1]}"
            component += f"\"{obj}\", "
        component = component.rstrip(", ")
        self._components.append(component)
    
    def from_type(self, *types: EObjectType):
        """
        Generates a sequence of objects from the specified types. One or many object types can be specified.
        :param types: Wwise object types.
        """
        component = "from type "
        for etype in types:
            component += f"{etype.get_type_name()}, "
        component = component.rstrip(", ")
        self._components.append(component)
    
    def from_query(self, query: ProjectPath | GUID):
        """
        Generates a sequence of objects from the result of executing a Wwise Query defined in the Query Editor.
        :param query: The project path or GUID of a Query object in Wwise.
        """
        self._components.append(f"from query \"{query}\"")
    
    def from_search(self, text: str):
        """
        Generates a sequence of objects from the result of a text search on the whole project.
        :param text: Each word specified in the search's text are matched to words inside the project objects.
        """
        self._components.append(f"from search \"{text}\"")
    
    def from_project(self):
        """Generates a sequence of objects including all objects of the project."""
        self._components.append(f"from project")
    
    def skip(self, count: int):
        """
        Bypasses a specified number of objects in the input sequence and then returns the remaining objects.
        :param count: The number of elements to skip.
        """
        self._components.append(f"skip {count}")
    
    def take(self, count: int):
        """
        Returns a specified number of contiguous objects from the start of the input sequence.
        :param count: The number of elements to take.
        """
        self._components.append(f"skip {count}")
    
    def distinct(self):
        """
        Returns only the distinct objects of the input sequence. Duplicated object entries will be reduced to one
        entry, resulting in each object being unique in the sequence.
        """
        self._components.append(f"distinct")
    
    def select(self, *expressions_or_property_names: EWaqlSelectExpression | str):
        """
        Returns an object or list of objects based on the specified expressions and Names.
        :param expressions_or_property_names: The expressions and/or property names to use.
        """
        component = "select "
        for ex_or_prop in expressions_or_property_names:
            component += f"{ex_or_prop}, "
        component = component.rstrip(", ")
        self._components.append(component)
    
    def where(self):
        """
        Adds "where" to your query. Use this in combination with **expression**, **and_expression**, **or_expression**,
        **open_bracket**, and **close_bracket** to create either simple or compound expressions.
        """
        self._components.append("where")
    
    def expression(self, property_name: str, logic_operator: EWaqlLogicalOperator | str,
                   value_or_ref_or_regex: bool | int | float | str | tuple[
                       EObjectType, Name] | ProjectPath | GUID | RegexPattern):
        """
        Filters objects of the input sequence by rejecting objects that don't match the specified expression.
        :param property_name: The name of the property to evaluate.
        :param logic_operator: The bool operator to use. Supported operators: `=`, `!=`, `<`, `<=`, `>`, `>=`, `:`.
        :param value_or_ref_or_regex: The value or the reference to use in the evaluation. Regex IS supported, but only
                                      when passing a `RegexPattern` (or `re.Pattern`) object.
        """
        if logic_operator == "==":  # Fixing common user mistake :)
            logic_operator = EWaqlLogicalOperator.EQUAL
        if str(logic_operator) not in EWaqlLogicalOperator:
            raise ValueError(f"Invalid logic operator for WAQL: '{logic_operator}'")
        match value_or_ref_or_regex:
            case str() | ProjectPath() | GUID():
                value_or_ref_or_regex = f"\"{value_or_ref_or_regex}\""
            case tuple():
                value_or_ref_or_regex = f"\"{value_or_ref_or_regex[0].get_type_name()}:{value_or_ref_or_regex[1]}\""
            case RegexPattern():
                value_or_ref_or_regex = f"/{value_or_ref_or_regex}/"
        self._components.append(f"{property_name} {logic_operator} {value_or_ref_or_regex}")
    
    def and_operator(self):
        """
        Append **and** to your query. Use this in combination with **expression**, **and_expression**, **or_expression**,
        **open_bracket**, and **close_bracket** to create compound expressions.
        """
        self._components.append("and")
    
    def or_operator(self):
        """
        Append **or** to your query. Use this in combination with **expression**, **and_expression**, **or_expression**,
        **open_bracket**, and **close_bracket** to create compound expressions.
        """
        self._components.append("or")
    
    def open_bracket(self):
        """
        Insert an open round bracket **("(")** into your query. Use this in combination with **expression**,
        **and_expression**, **or_expression**, **open_bracket**, and **close_bracket** to create compound expressions.
        """
        self._components.append("(")
    
    def close_bracket(self):
        """
        Insert an open round bracket **(")")** into your query. Use this in combination with **expression**,
        **and_expression**, **or_expression**, **open_bracket**, and **close_bracket** to create compound expressions.
        """
        self._components.append(")")
